/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse;

import static org.fortiss.af3.exploration.alg.service.ExplorationService.getService;
import static org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext.DSE_INTERNAL;
import static org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory.createSingleExplorationSolution;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.instantiatedmapping.InstantiatedTaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IExplorationSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.IScheduleSolution;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution.StrictTTSchedule;
import org.fortiss.af3.exploration.alg.service.ExplorationTransformationInputs;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.af3.exploration.model.SuperSetMap;
import org.fortiss.af3.exploration.model.solutions.ExplorationResult;
import org.fortiss.af3.exploration.model.solutions.ExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SingleExplorationSolution;
import org.fortiss.af3.exploration.model.solutions.SolutionState;
import org.fortiss.af3.exploration.model.solutions.SolutionsFactory;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.opt4j.core.Value;

import com.google.common.collect.ClassToInstanceMap;
import com.google.common.collect.MutableClassToInstanceMap;

/**
 * Result of a DSE run, i.e. the {@link IExplorationSolution}s and {@link SolutionQuantification}s
 * for all pareto-optimal {@link IExplorationSolution}s.
 * 
 * @author barner
 */
public class CompositeExplorationSolution<P extends Phenotype> {
	/**
	 * Map to store the {@link SolutionQuantification} for all pareto-optimal
	 * {@link IExplorationSolution}s of an exploration run.
	 */
	private Map<IExplorationSolution<?>, SolutionQuantification> explorationResultMap =
			new HashMap<IExplorationSolution<?>, SolutionQuantification>();

	/** Number of completed iterations that produced this {@link CompositeExplorationSolution}. */
	private int completedIterations;

	/**
	 * Adds the {@link SolutionQuantification} for a given {@link IExplorationSolution}.
	 */
	public void addExplorationSolution(IExplorationSolution<?> solution,
			SolutionQuantification result) {
		explorationResultMap.put(solution, result);
	}

	/**
	 * Deletes a {@link IExplorationSolution} and its according {@link SolutionQuantification} from
	 * the map containing the results.
	 * This method is useful for removing, e.g., infeasible results.
	 */
	public void deleteSolutionQuantification(IExplorationSolution<?> solution) {
		explorationResultMap.remove(solution);
	}

	/**
	 * Returns the Map of exploration solutions in the form {@link IExplorationSolution},
	 * {@link SolutionQuantification}. This method is intended to provide intermediate or final
	 * solution to the user.
	 */
	public Map<IExplorationSolution<?>, SolutionQuantification> getSolutionQuantification() {
		return explorationResultMap;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String rval =
				"After " + completedIterations +
						" completed iterations, the exploration result contains " +
						explorationResultMap.size() + " individual(s) (= archive size).\n\n";

		for(IExplorationSolution<?> solution : explorationResultMap.keySet()) {
			IScheduleSolution<?> scheduleSolution = solution.getSubSolution(StrictTTSchedule.class);
			InstantiatedTaskMappingEncoding<?> mappingSolution =
					solution.getSubSolution(InstantiatedTaskMappingEncoding.class);
			rval += "------------------------------------------------\n";
			rval += "Encoding (genotype):\n" + mappingSolution + "\n";
			rval += "Solution (phenotype):\n" + scheduleSolution + "\n";
		}

		return rval;
	}

	/** Creates a copy of the ExplorationResult object this operator is applied to. */
	public CompositeExplorationSolution<P> copy() {
		CompositeExplorationSolution<P> duplicatedExplorationResult =
				new CompositeExplorationSolution<P>();

		for(IExplorationSolution<?> solution : explorationResultMap.keySet()) {
			duplicatedExplorationResult.addExplorationSolution(solution,
					explorationResultMap.get(solution));
		}

		duplicatedExplorationResult.setCompletedIterations(completedIterations);

		return duplicatedExplorationResult;
	}

	/** Records the number of completed iterations. */
	public void setCompletedIterations(int completedIterations) {
		this.completedIterations = completedIterations;
	}

	/** Returns the number of completed iterations. */
	public int getCompletedIterations() {
		return completedIterations;
	}

	/**
	 * Returns {@code this} {@link CompositeExplorationSolution} as a {@link ExplorationSolution} .
	 */
	@SuppressWarnings("unchecked")
	public <T, S extends IModelElement> ExplorationSolution convertToExplorationSolution()
			throws Exception {
		ExplorationSolution expSol =
				org.fortiss.af3.exploration.util.DesignSpaceExplorationModelElementFactory
						.createExplorationSolution();
		for(Entry<IExplorationSolution<?>, SolutionQuantification> singleSolutionEntry : explorationResultMap
				.entrySet()) {
			SingleExplorationSolution singleExpSolution = createSingleExplorationSolution();
			IExplorationSolution<?> singleSolution = singleSolutionEntry.getKey();

			for(Entry<ExplorationTarget<?>, Value<?>> solEntry : singleSolutionEntry.getValue()
					.entrySet()) {
				if(singleExpSolution.containsKey(solEntry.getKey())) {
					throw new Exception(
							"Could not convert to the ExplorationResult to an ExplorationSolutionSet. The exist multiple results for the ExplorationTarget " +
									solEntry.getKey().getName() +
									" where there should be only one.");
				}
				// TODO: this is unsafe, the SolutionQuantification class should be either
				// eliminated, or adapted such that we obtain the type safety that is also aimed
				// at in the EMM.
				@SuppressWarnings("rawtypes") ExplorationResult res =
						SolutionsFactory.eINSTANCE.createExplorationResult();
				res.setResult(solEntry.getValue().getDouble());
				singleExpSolution.put((ExplorationTarget<?>)solEntry.getKey(), res);
			}

			ClassToInstanceMap<Phenotype> internalSolutions = MutableClassToInstanceMap.create();
			for(Class<? extends Phenotype> solutionType : singleSolution.getAllSubSolutionClasses()) {
				internalSolutions.put(solutionType, singleSolution.getSubSolution(solutionType));
			}

			// FIXME: HACK: Currently, only the Deployment exploration is used, whereas the solution
			// transformation framework is non-functional at the moment.
			Collection<Class<?>> outputTypes = new ArrayList<>();
			// In a later stage, we would pass a list of generated "generatable" SuperSets derived
			// from the Phenotypes. This is the inverse setup of the model-oriented transformations
			// where we request desired types. Here, we'd calculate all types that can be produced.
			outputTypes.add(SuperSetMap.class);
			ExplorationTransformationInputs inputs =
					new ExplorationTransformationInputs(null, internalSolutions);
			Map<Class<EObject>, ?> solutionModels =
					getService(IExplorationTransformationService.class).getTransformedModels(
							DSE_INTERNAL, null, inputs, outputTypes, "");
			singleExpSolution.setSolutionSets((SuperSetMap)solutionModels.get(SuperSetMap.class));

			// Invalid solutions are filtered before.
			singleExpSolution.setSolutionState(SolutionState.OPTIMIZED);

			expSol.getSolutions().add(singleExpSolution);
		}

		if(expSol.getSolutions().isEmpty()) {
			expSol.setSolutionState(SolutionState.UNKNOWN);
		} else {
			expSol.setSolutionState(SolutionState.OPTIMIZED);
		}

		return expSol;
	}
}
