/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.selection.IOperatorSelector;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.operator.selection.OperatorSelectorPredicate;
import org.opt4j.core.Genotype;
import org.opt4j.core.optimizer.Operator;
import org.opt4j.operators.AbstractGenericOperator.OperatorPredicate;

import com.google.inject.Binder;
import com.google.inject.TypeLiteral;
import com.google.inject.multibindings.MapBinder;

/**
 * This module groups mutation operators that affect the mapping of tasks to execution units.
 * 
 * Base class for mutate operator module that enables the registration of pre-constructed operators,
 * which thus can be constructed using a non-default constructor.
 * 
 * @author barner
 */
public class OperatorUtil {

	/**
	 * Registers a pre-constructed operator instance with the Opt4J framework, i.e. an operator that
	 * can be built using a non-default constructor and that thus can be based on external state.
	 */
	public static <O extends Operator<?>> void addOperatorInstance(O operator,
			IOperatorSelector operatorSelector, Binder binder, TypeLiteral<O> typeLiteral) {
		addOperatorMapBinding(binder, operator, typeLiteral, new OperatorSelectorPredicate(
				operator, operatorSelector));
	}

	/**
	 * Registers a pre-constructed operator instance with the Opt4J framework, i.e. an operator that
	 * can be built using a non-default constructor and that thus can be based on external state.
	 * The operator will be only applied to the given {@link Genotype}.
	 */
	public static <G extends Genotype, O extends Operator<?>> void addOperatorInstance(
			Class<G> genotypeType, O operator, IOperatorSelector operatorSelector, Binder binder,
			TypeLiteral<O> typeLiteral) {
		addOperatorMapBinding(binder, operator, typeLiteral, new OperatorSelectorPredicate(
				operator, new ComposableOperatorClassPredicate(genotypeType), operatorSelector));
	}

	/** Registers a pre-constructed operator and predicate instance with the Opt4J framework. */
	private static <O extends Operator<?>> void addOperatorMapBinding(Binder binder, O operator,
			TypeLiteral<O> typeLiteral, OperatorSelectorPredicate predicate) {
		// This is inspired from org.opt4j.operators.OperatorModule.addOperator(), which does not
		// allow to pass parameters to the constructor of the operators. Hence, a pre-constructed
		// mutate operator is bound here (using toInstance() instead of to()).
		MapBinder<OperatorPredicate, O> map =
				MapBinder.newMapBinder(binder, new TypeLiteral<OperatorPredicate>() { // empty
						}, typeLiteral);
		map.addBinding(predicate).toInstance(operator);
	}
}
