/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.backend.opt4j.solution;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.partitionmapping.Partition;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.encoding.taskmapping.TaskMappingEntry;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.ComposablePhenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ICommunicationResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IExecutionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IGatewayUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IRequestAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.IResourceAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITaskAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.ITransmissionUnitAdapter;
import org.fortiss.af3.exploration.alg.dse.sysmodel.comm.Message;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEncoding;
import org.fortiss.af3.exploration.alg.dse.sysmodel.mapping.IMappingEntry;
import org.fortiss.af3.exploration.alg.dse.sysmodel.sched.ScheduleRuntimeException;

/**
 * Time-triggered schedule encoding the temporal execution order of a given
 * {@link TaskMappingEncoding}.
 * 
 * @author huang
 */

public class StrictTTSchedule<S extends TaskMappingEntry<ITaskAdapter<?>, Partition>, T extends TaskMappingEncoding<ITaskAdapter<?>, S>>
		implements IScheduleSolution<StrictTTSchedule<S, T>>, ComposablePhenotype<T>,
		ArchExpSubSolution<T> {
	/**
	 * Amount (of time) which {@link TimeSlot}s are allowed to overlap. Here due to numerical
	 * imprecision.
	 */
	// TODO: This should be better placed in the scheduler as some sort of "jitter".
	public static double SLOT_OVERLAP_TOLERANCE = 1e-6;

	/**
	 * {@link HashMap}-based Schedule of a single resource = map: time slot -> request allocated to
	 * that time slot
	 */
	public static class ResourceSchedule extends TreeMap<TimeSlot, IRequestAdapter<?>> {
		/**
		 * Indicates whether the temporal order within a {@link ResourceSchedule} is kept. This is
		 * done by evaluating the predecessor field of {@link TimeSlot}s
		 */
		public boolean validateResourceSchedule() {
			for(TimeSlot slot : keySet()) {
				for(TimeSlot pred : slot.getPredecessors()) {
					if(pred.getResource() == slot.getResource() &&
							slot.getStartTime() < pred.getEndTime()) {
						return false;
					}
				}
			}
			return true;
		}
	}

	/**
	 * Schedule data structure = map: resource -> scheduling entries that resource
	 */
	private Map<IResourceAdapter<?>, ResourceSchedule> schedule;

	/**
	 * Associates the {@link IMappingEntry}s with their {@link TimeSlot}s. Essentially, a different
	 * perspective onto the {#link schedule}.
	 */
	private Map<S, TimeSlot> timeSlotsOfMappingEntries;

	// TODO: Check if there is an Opt4J mechanism to obtain the genotype for a phenotype if this is
	// required (e.g., in an evaluator)
	/** Maintain a reference to the encoding. */
	private T encoding;

	/**
	 * Constructor that creates a time-triggered schedule for a given encoding (mapping) and
	 * optionally directly performs the scheduling.
	 */
	public StrictTTSchedule(T encoding) {
		this.encoding = encoding;
		schedule = new HashMap<IResourceAdapter<?>, ResourceSchedule>();
		timeSlotsOfMappingEntries = new HashMap<S, TimeSlot>();
	}

	/** Copy constructor. */
	public StrictTTSchedule(StrictTTSchedule<S, T> schedule) {
		this.encoding = schedule.encoding;
		this.schedule = new HashMap<IResourceAdapter<?>, ResourceSchedule>();
		timeSlotsOfMappingEntries = new HashMap<S, TimeSlot>();

		for(Entry<IResourceAdapter<?>, ResourceSchedule> resSched : schedule.schedule.entrySet()) {
			ResourceSchedule newSched = new ResourceSchedule();
			newSched.putAll(resSched.getValue());
			this.schedule.put(resSched.getKey(), newSched);
		}

		timeSlotsOfMappingEntries.putAll(schedule.timeSlotsOfMappingEntries);
	}

	/**
	 * Performs a validation of this schedule. Returns {@code true} if the following conditions
	 * hold:
	 * <li>The number of allocations defined in the mapping matches the number of scheduled
	 * components.</li>
	 * <li>No overlap of {@link TimeSlot}s within the schedule of each
	 * {@link IResourceAdapter}.</li>
	 */
	public boolean validateSchedule(boolean enableScheduleException) {
		if(validateAllocationNumberInSchedule(enableScheduleException) &&
				validateNoTimeSlotOverlaps(enableScheduleException) &&
				validateCausalityOfTasks(enableScheduleException)) {
			return true;
		}

		return false;
	}

	/**
	 * Validates whether the number of {@link TimeSlot}s representing a
	 * {@link ITaskAdapter} matches the number of allocations defined in the
	 * {@link IMappingEncoding} from which this {@link StrictTTSchedule} is derived.
	 * This check is performed for each {@link ITaskAdapter} in the
	 * {@link IMappingEncoding}.
	 */
	private boolean validateAllocationNumberInSchedule(boolean enableScheduleException) {
		for(ITaskAdapter<?> comp : encoding.getRequesters()) {
			int compEncNum = encoding.getMappingEntriesOf(comp).size();
			int compSchedNum = 0;
			for(IResourceAdapter<?> resourceAdapter : getResources()) {
				for(IRequestAdapter<?> schedComp : getResourceSchedule(resourceAdapter).values()) {
					if(schedComp == comp) {
						compSchedNum++;
					}
				}
			}
			if(compEncNum != compSchedNum && encoding.getRequesters().contains(comp)) {
				System.out.println("The number of scheduled Components of " + comp.getName() +
						" does not match the number defined in the Component->ExecutionUnit mapping.");
				if(enableScheduleException) {
					throw new ScheduleRuntimeException(this);
				}
				return false;
			}
		}
		return true;
	}

	/**
	 * Validates for each {@link ResourceSchedule} that no {@link TimeSlot}s have temporal overlaps
	 * with another {@link TimeSlot} of that {@link ResourceSchedule}.
	 */
	private boolean validateNoTimeSlotOverlaps(boolean enableScheduleException) {
		for(Map.Entry<IResourceAdapter<?>, ResourceSchedule> resourceSchedule : schedule
				.entrySet()) {
			for(TimeSlot currentTimeSlot : resourceSchedule.getValue().keySet()) {
				TimeSlot nextTimeSlot = resourceSchedule.getValue().higherKey(currentTimeSlot);
				if(nextTimeSlot != null &&
						currentTimeSlot.getEndTime() > nextTimeSlot.getStartTime() +
								SLOT_OVERLAP_TOLERANCE * nextTimeSlot.getStartTime()) {
					System.out.println("The time slots " + currentTimeSlot + " and " +
							nextTimeSlot + " overlap in the schedule of the resource " +
							resourceSchedule.getKey().getName() + ". The schedule is not valid.");
					if(enableScheduleException) {
						throw new ScheduleRuntimeException(this);
					}
					return false;
				}
			}
		}

		return true;
	}

	/**
	 * Predicate if this {@link StrictTTSchedule} is causally valid by the order of its
	 * {@link TimeSlot}s.
	 * <b>NOTE:</b> Only those {@link TimeSlot} which that are contained in this
	 * {@link StrictTTSchedule} and for which the predecessor {@link TimeSlot}s are defined are
	 * validated.
	 */
	private boolean validateCausalityOfTasks(boolean enableScheduleException) {
		for(Map.Entry<IResourceAdapter<?>, ResourceSchedule> resourceSchedule : schedule
				.entrySet()) {
			for(TimeSlot currentTimeSlot : resourceSchedule.getValue().keySet()) {
				for(TimeSlot predecessorSlot : currentTimeSlot.getPredecessors()) {
					if(currentTimeSlot.getStartTime() < predecessorSlot.getEndTime() -
							StrictTTSchedule.SLOT_OVERLAP_TOLERANCE) {
						System.out.println("(Temporal) Causality of the time slots " +
								predecessorSlot + " (predecessor, resource: " +
								predecessorSlot.getResource().getName() + ") and " +
								currentTimeSlot + "(successor, resource: " +
								currentTimeSlot.getResource().getName() +
								") violated. The schedule is not valid.");
						if(enableScheduleException) {
							throw new ScheduleRuntimeException(this);
						}
						return false;
					}
				}
			}
		}

		return true;
	}

	/**
	 * Creates a {@link TimeSlot} for a voter and inserts it into the schedule (voters are software
	 * components that can only be allocated to {@link IExecutionUnitAdapter}s.
	 */
	public TimeSlot createVoterSlot(S deployedRequest, IExecutionUnitAdapter<?> deploymentTarget,
			double start, double end, long iteration) {
		TimeSlot slot = createTimeSlot(deployedRequest, deploymentTarget, start, end, iteration);
		slot.setVoter(true);
		return slot;
	}

	/** Registers a {@link TimeSlot} with the appropriate schedule. */
	public TimeSlot registerTimeSlot(TimeSlot slot, S deployedRequest,
			IResourceAdapter<?> resource) {
		// Create a new TimeSlot and eventually add a new ResourceSchedule, if none exists yet for
		// the deployment target.
		if(schedule.get(resource) == null) {
			ResourceSchedule resourceSchedule = new ResourceSchedule();
			schedule.put(resource, resourceSchedule);
		}

		// Separate the registration within the schedule by the mapping entry. If it is non-null,
		// the mapping entry represents one instance of a deployable component and is registered
		// within an appropriate ResourceSchedule and task mapping map. Otherwise, just register the
		// time slot with no associated deployable Component within the appropriate
		// ReosurceSchedule; it is most likely a communication resource allocated to a message.
		if(deployedRequest != null) {
			schedule.get(resource).put(slot, deployedRequest.getSource());
			timeSlotsOfMappingEntries.put(deployedRequest, slot);
		} else {
			schedule.get(resource).put(slot, null);
		}

		return slot;
	}

	/**
	 * Creates a {@link TimeSlot} with an iteration number and inserts it into the appropriate
	 * schedule.
	 */
	public TimeSlot createTimeSlot(S deployedRequest, IResourceAdapter<?> resource, double start,
			double end, long iteration) {
		TimeSlot slot = new TimeSlot(deployedRequest, resource, start, end, (int)iteration, false);
		return registerTimeSlot(slot, deployedRequest, resource);
	}

	/**
	 * Creates a {@link TimeSlot} with an iteration number and inserts it into the appropriate
	 * schedule. Additionally, this method sets the predecessor {@link TimeSlot}s, which can be used
	 * to verify the causal correctness.
	 */
	public TimeSlot createTimeSlot(S deployedRequest, IResourceAdapter<?> resource,
			Collection<TimeSlot> predessorSlots, double start, double end, long iteration) {
		TimeSlot slot = new TimeSlot(deployedRequest, resource, predessorSlots, start, end,
				(int)iteration, false);
		return registerTimeSlot(slot, deployedRequest, resource);
	}

	/** Creates a {@link TimeSlot} specifically for {@link Message}s. */
	public TimeSlot createMessageSlot(Message message, ICommunicationResourceAdapter<?> resource,
			double start, double end) {
		TimeSlot slot = new TimeSlot(message, resource, start, end);
		// TODO: Modify the Message class to contain a reference to the *original* sender.
		return registerTimeSlot(slot, null, resource);
	}

	/** Returns the resources whose schedule is contained in this {@link StrictTTSchedule}. */
	public Set<IResourceAdapter<?>> getResources() {
		return schedule.keySet();
	}

	/** Returns the {@link ResourceSchedule} for a given resource */
	public ResourceSchedule getResourceSchedule(IResourceAdapter<?> resource) {
		return schedule.get(resource);
	}

	/** {@inheritDoc} */
	@Override
	public TimeSlot getTimeSlotOf(IMappingEntry<?, ?> entry) {
		return timeSlotsOfMappingEntries.get(entry);
	}

	/**
	 * Return the earliest start time of all time slots that are allocated for the given request.
	 */
	public double getFirstStartTime(IRequestAdapter<?> request, long iteration) {

		double time = -1;
		for(ResourceSchedule resourceSchedule : schedule.values()) {
			for(TimeSlot t : resourceSchedule.keySet()) {
				if(resourceSchedule.get(t) == request) {
					if(t.getIteration() == iteration) {
						if(time < 0 || time > t.getStartTime()) {
							time = t.getStartTime();
						}
					}
				}
			}
		}
		return time;
	}

	/** Return the latest end time of all time slots that are allocated for the request. */
	public double getLastEndTime(IRequestAdapter<?> request, long iteration) {

		double time = -1;
		for(ResourceSchedule resourceSchedule : schedule.values()) {
			for(TimeSlot t : resourceSchedule.keySet()) {
				if(resourceSchedule.get(t) == request) {
					if(t.getIteration() == iteration) {
						if(time < 0 || (time > 0 && time < t.getEndTime())) {
							time = t.getEndTime();
						}
					}
				}
			}
		}
		return time;
	}

	/** Formats the string representation of a resource */
	private String getResourceName(IResourceAdapter<?> resource) {
		String str = resource.getName();
		if(!(resource instanceof ITransmissionUnitAdapter ||
				resource instanceof IGatewayUnitAdapter)) {
			// FIXME: Use the name of the partition.
			str += "@" + resource
					.toString();/* systemModelAdapter.getHardwareResource(resource).getName(); */
		}
		return str;
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		String str = "";
		for(IResourceAdapter<?> resource : schedule.keySet()) {
			str += getResourceName(resource);
			str += ":";
			for(TimeSlot slot : schedule.get(resource).keySet()) {
				str += "[";
				if(slot.isVoter()) {
					str += "voter";
				} else if(slot.getDeployedRequester() != null) {
					str += slot.getDeployedRequester().getSource().getName();
				} else if(slot.getDeployedMessage() != null) {
					str += slot.getDeployedMessage().getSignalAdapter().getName();
				} else {
					str += "<>";
				}
				str += " @ " + slot + "], ";
			}
			str += "\n";
		}
		return str;
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public Class<? extends ComposablePhenotype<?>> getPTypeToRegister() {
		return (Class<? extends ComposablePhenotype<?>>)StrictTTSchedule.class;
	}
}
