/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.evaluator;

import java.util.Collection;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.FlatPhenotypeMap;
import org.fortiss.af3.exploration.alg.dse.backend.opt4j.extensions.compositegene.phenotype.Phenotype;
import org.fortiss.af3.exploration.alg.dse.sysmodel.arch.SystemModelAdapter;
import org.fortiss.af3.exploration.alg.exception.ExplorationException;
import org.fortiss.af3.exploration.alg.service.IExplorationTargetEvaluator;
import org.fortiss.af3.exploration.model.ExplorationConstraint;
import org.fortiss.af3.exploration.model.ExplorationTarget;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.opt4j.core.DoubleValue;
import org.opt4j.core.Value;
import org.opt4j.core.problem.Evaluator;

import com.google.common.collect.ClassToInstanceMap;

/**
 * Base class for {@link Evaluator}s of {@link ExplorationConstraint}s..
 * 
 * @author barner
 */
public abstract class ExplorationTargetEvaluatorBase<P, RT, T extends ExplorationTarget<RT>>
		implements IExplorationTargetEvaluator<P, RT, T> {

	/** Representation of system model */
	protected SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter;

	/** Exploration goal evaluated by this evaluator */
	protected T explorationTarget;

	/** Prototype constructor. */
	protected ExplorationTargetEvaluatorBase() {
	}

	/**
	 * Creates an evaluator for a given exploration goal
	 * 
	 * @throws Exception
	 *             if the input validation fails (has to be implemented by the concrete
	 *             {@link ExplorationTarget} evaluators.
	 */
	protected ExplorationTargetEvaluatorBase(
			SystemModelAdapter<?, ?, ?, ?, ?, ?, ?> systemModelAdapter, T optTarget)
			throws Exception {
		this.systemModelAdapter = systemModelAdapter;
		this.explorationTarget = optTarget;

		validateInputs();
	}

	/** Wrapper method for the original evaluate function. */
	public Value<?> evaluate(P phenotype, FlatPhenotypeMap<Phenotype> phenotypeTypeMap,
			ClassToInstanceMap<EObject> transformedModels) {
		if(phenotype == null) {
			return new DoubleValue(Double.MAX_VALUE);
		}
		// TODO #checking: Add an check here to ensure that the evaluators don't mess around with
		// the encodings.
		Value<?> evalResult = new DoubleValue(Double.MAX_VALUE);
		try {
			evalResult = evaluateGoal(phenotype, phenotypeTypeMap, transformedModels);
		} catch(Exception e) {
			// The Opt4J framework only allows us to throw runtime exceptions.
			throw new RuntimeException(new ExplorationException("The evaluatior " +
					this.toString() + " failed to evaluate the ExplorationTarget " +
					explorationTarget.getName() + ".", e));
		}
		return evalResult;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return getClass().getSimpleName();
	}

	// TODO: Add the introspection information.
	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
