/*-------------------------------------------------------------------------+
| Copyright 2016 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.exploration.alg.dse.modeltransformation.partition;

import static org.fortiss.af3.allocation.utils.AllocationModelElementFactory.createAllocationTableCollection;
import static org.fortiss.af3.allocation.utils.AllocationUtils.addAllocationEntry;
import static org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule.TransformationState.INITIALIZING;
import static org.fortiss.af3.exploration.alg.util.TransformationUtils.getArchitectureOfAllocSuperSet;
import static org.fortiss.tooling.common.util.LambdaUtils.getFirst;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.af3.allocation.model.AllocationTableCollection;
import org.fortiss.af3.component.model.ComponentArchitecture;
import org.fortiss.af3.deployment.model.Deployment;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.ITransformationModule;
import org.fortiss.af3.exploration.alg.dse.modeltransformation.base.TransformationModuleBase;
import org.fortiss.af3.exploration.alg.guava.SubClassToInstanceMap;
import org.fortiss.af3.exploration.alg.service.ExplorationTransformationInputs;
import org.fortiss.af3.exploration.alg.service.IExplorationTransformationService.TransformationContext;
import org.fortiss.af3.exploration.dsl_v2.model.expression.SuperSet;
import org.fortiss.af3.partition.model.Partition;
import org.fortiss.af3.partition.model.PartitionArchitecture;
import org.fortiss.af3.partition.model.allocation.PartitionToExecutionUnitAllocationEntry;
import org.fortiss.af3.partition.model.allocation.PartitionToExecutionUnitAllocationTable;
import org.fortiss.af3.partition.model.allocation.TaskToPartitionAllocationEntry;
import org.fortiss.af3.partition.model.allocation.TaskToPartitionAllocationTable;
import org.fortiss.af3.partition.util.PartitionModelElementFactory;
import org.fortiss.af3.platform.model.PlatformArchitecture;
import org.fortiss.af3.project.model.FileProject;
import org.fortiss.af3.project.utils.ProjectUtils;
import org.fortiss.af3.task.model.TaskArchitecture;
import org.fortiss.tooling.base.model.element.IHierarchicElement;
import org.fortiss.tooling.common.util.LambdaUtils;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * Initializes a {@link Deployment} by creating a new model and setting its references to the given
 * {@link ComponentArchitecture} and {@link PlatformArchitecture}.
 * 
 * @author diewald
 */
public class PartitionTransformer extends TransformationModuleBase<AllocationTableCollection> {
	/** Input {@link TaskArchitecture} passed to the DSE. */
	private TaskArchitecture taskArchitecture;
	/** Input {@link PlatformArchitecture} passed to the DSE. */
	private PlatformArchitecture platfromArchitecture;
	/** Generated {@link PartitionArchitecture} by the DSE. */
	private PartitionArchitecture partitionArchitecture;
	/**
	 * Generated {@link AllocationTableCollection} containing task-to-partition and
	 * partition-to-execution-unit allocation tables.
	 */
	private AllocationTableCollection allocTableCollection;

	/** {@inheritDoc} */
	@Override
	public <N extends EObject, S extends EObject> AllocationTableCollection transform(
			TransformationState state, TransformationContext context,
			ExplorationTransformationInputs inputs, SubClassToInstanceMap<N> transformedModels,
			EObject outputModel) {
		platfromArchitecture =
				getArchitectureOfAllocSuperSet(
						inputs.getSuperSet(PartitionToExecutionUnitAllocationEntry.class),
						PlatformArchitecture.class);
		taskArchitecture =
				getArchitectureOfAllocSuperSet(
						inputs.getSuperSet(TaskToPartitionAllocationEntry.class),
						TaskArchitecture.class);

		SuperSet<PartitionToExecutionUnitAllocationEntry> partToExU =
				inputs.getSuperSet(PartitionToExecutionUnitAllocationEntry.class);
		// Get a random entry to retrieve the model context.
		PartitionToExecutionUnitAllocationEntry entry =
				LambdaUtils.getFirst(partToExU.getEntries()).get();
		FileProject fp =
				ProjectUtils.getFileProject(LambdaUtils.getFirst(entry.getTargetElements()).get());
		ITopLevelElement modelContext = IPersistencyService.getInstance().getTopLevelElementFor(fp);

		// TODO: The following step should be performed in a separate module.
		createPartitionArchitecture(inputs.getSuperSet(Partition.class), modelContext);
		modelContext.runAsNonDirtyingCommand(() -> fp.getRootElements().add(partitionArchitecture));

		// Create the allocations.
		allocTableCollection = createAllocationTableCollection("Generated Allocations");
		TaskToPartitionAllocationTable taskExUTable =
				createTaskToPartitionAllocationTable(inputs
						.getSuperSet(TaskToPartitionAllocationEntry.class));
		modelContext.runAsNonDirtyingCommand(() -> allocTableCollection.getContainedElements().add(
				taskExUTable));
		PartitionToExecutionUnitAllocationTable partExUTable =
				createPartitionToExecutionUnitAllocationTable(inputs
						.getSuperSet(PartitionToExecutionUnitAllocationEntry.class));
		modelContext.runAsNonDirtyingCommand(() -> allocTableCollection.getContainedElements().add(
				partExUTable));

		return allocTableCollection;
	}

	/** Creates a {@link PartitionArchitecture} from the entries in the given {@link SuperSet}. */
	private void createPartitionArchitecture(SuperSet<Partition> partitions,
			ITopLevelElement modelContext) {
		partitionArchitecture =
				PartitionModelElementFactory
						.createPartitionArchitecture("Generated Partition Architecture");
		EList<IHierarchicElement> partitionList = partitionArchitecture.getContainedElements();
		modelContext.runAsNonDirtyingCommand(() -> partitionList.addAll(partitions.getEntries()));
	}

	/**
	 * Creates a {@link TaskToPartitionAllocationTable} from the corresponding entries in
	 * the given {@link SuperSet}.
	 */
	private TaskToPartitionAllocationTable createTaskToPartitionAllocationTable(
			SuperSet<TaskToPartitionAllocationEntry> taskExecUnitAllocs) {
		TaskToPartitionAllocationTable table =
				org.fortiss.af3.partition.util.PartitionModelElementFactory
						.createTaskToPartitionAllocationTable("Generated Task->Part Table");
		table.setSourceView(taskArchitecture);
		table.setTargetView(partitionArchitecture);

		for(TaskToPartitionAllocationEntry entry : taskExecUnitAllocs.getEntries()) {
			Partition af3Partition = (Partition)entry.getTargetElement();
			// TODO: We assume single Tasks in this place.
			addAllocationEntry(table, TaskToPartitionAllocationEntry.class,
					getFirst(entry.getSourceElements()).get(), af3Partition);
		}
		return table;
	}

	/**
	 * Creates a {@link PartitionToExecutionUnitAllocationTable} from the corresponding entries in
	 * the given {@link SuperSet}.
	 */
	private PartitionToExecutionUnitAllocationTable createPartitionToExecutionUnitAllocationTable(
			SuperSet<PartitionToExecutionUnitAllocationEntry> partMapping) {
		PartitionToExecutionUnitAllocationTable table =
				PartitionModelElementFactory
						.createPartitionToExecutionUnitAllocationTable("Generated Part->ExU Table");
		table.setSourceView(partitionArchitecture);
		table.setTargetView(platfromArchitecture);

		for(PartitionToExecutionUnitAllocationEntry entry : partMapping.getEntries()) {
			Partition af3Partition = (Partition)entry.getSourceElement();
			// TODO: We assume single ExecutionUnits as allocation targets in this place.
			addAllocationEntry(table, PartitionToExecutionUnitAllocationEntry.class, af3Partition,
					getFirst(entry.getTargetElements()).get());
		}
		return table;
	}

	/** {@inheritDoc} */
	@Override
	public Class<AllocationTableCollection> getOutputType() {
		return AllocationTableCollection.class;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getRequiredTypes() {
		Collection<Class<?>> inputTypes = new HashSet<>();
		return inputTypes;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<?>> getOptionalTypes() {
		return Collections.emptySet();
	}

	/** {@inheritDoc} */
	@Override
	public Collection<Class<? extends ITransformationModule<?>>> getParentModules() {
		Collection<Class<? extends ITransformationModule<?>>> parentModules = new HashSet<>();
		parentModules.add(PartitionTransformationStrategy.class);
		return parentModules;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isOptional() {
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public TransformationState getActivationStage() {
		return INITIALIZING;
	}
}
